use Test::More tests => 17;
use_ok('Bundle::Zoidberg');
use_ok('Zoidberg');
use_ok('Zoidberg::Contractor');
use_ok('Zoidberg::DispatchTable');
use_ok('Zoidberg::Fish');
use_ok('Zoidberg::Fish::Commands');
use_ok('Zoidberg::Fish::Intel');
use_ok('Zoidberg::Fish::Log');
use_ok('Zoidberg::Fish::ReadLine');
use_ok('Zoidberg::PluginHash');
use_ok('Zoidberg::Shell');
use_ok('Zoidberg::StringParser');
use_ok('Zoidberg::Utils');
use_ok('Zoidberg::Utils::Error');
use_ok('Zoidberg::Utils::FileSystem');
use_ok('Zoidberg::Utils::GetOpt');
use_ok('Zoidberg::Utils::Output');

